@@ -153,8 +153,8 @@ META_ITEM *meta_getitembyextspec(META *m, char *spec, ssize_t specl,
    value itself. 
 
    The skip length can be obtained as follows: if the item has a length field,
-   (detected by len_size being non-zero), take its value and add the adjustment
-   given by len_adj, which can be positive, negative or zero.
+   (ie. len_size is non-zero), take its value and add the adjustment given by
+   len_adj, which can be positive, negative or zero.
 
    If there is no length field, and len_adj is zero or less, take the size
    of the enclosing block and add len_adj. If there is no length field but
@@ -260,11 +260,12 @@ static void debug_attr(META_SPC *s, META_ITEM *i, char *blk, char *p,
  * to its type, from raw data at val, vallen. */
 
 static META_AV *addavfromraw(META_AV **head, META_AV **tail,
-			 META_ITEM *i, char *p, META_ORD l, int copystrings)
+			     META_ITEM *i, char *p, META_ORD l, int copystrings)
 {
     META_AV *av;
 
-    av = (META_AV *)malloc(sizeof(META_AV)); if (!av) goto error;
+    av = (META_AV *)malloc(sizeof(META_AV)); 
+    if (!av) goto error;
     memset(av, 0, sizeof(META_AV));
 
     av->i = i;
@@ -281,7 +282,10 @@ static META_AV *addavfromraw(META_AV **head, META_AV **tail,
 	     * a temporary string as the source. */
 	    av->l = l; 
 	    av->p = (char *)malloc(l);
-	    if (!av->p) { free(av); goto error; }
+	    if (!av->p) { 
+		free(av); 
+		goto error; 
+	    }
 	    memcpy(av->p, p, av->l); 
 	    av->flags = AV_FREE_P; 
     }
@@ -293,7 +297,8 @@ static META_AV *addavfromraw(META_AV **head, META_AV **tail,
 
     return av;
 error:
-    meta_freeavlist(*head); *head = *tail = 0; 
+    meta_freeavlist(*head); 
+    *head = *tail = 0; 
     return 0;
 }
 
@@ -340,7 +345,8 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 			if (!i->nodec) {
 			    if (!addavfromraw(&avlist, &tail, i,
 					      blk + i->val_ofs, vallen,
-					      copystrings)) return 0;
+					      copystrings)) 
+				return 0;
 			}
 
 			/* Recurse into a subspace, if any */
@@ -358,9 +364,14 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 			}
 		}
 
-		/* If we terminated prematurely, show error and return */
-		if (i) { debug_fixed(i, blk, blksize); meta_freeavlist(avlist); 
-			 return 0; }
+		/* If we terminated prematurely, show error and return.
+		   This is only the case if a field crosses the block. */
+
+		if (i && i->len_ofs < blksize && i->val_ofs < blksize) {
+		    debug_fixed(i, blk, blksize); 
+		    meta_freeavlist(avlist); 
+		    return 0; 
+		}
 
 		/* Return the head of the created list */
 		if (rettail) *rettail = tail;
@@ -372,7 +383,7 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 	   actally got assigned to. Oh well. */
 	skiplen = 0;
 
-	/* Attribute space. You've just got to love C's logical operators... */
+	/* Attribute space. You've just got to love shortcut boolean eval... */
 	for(p = blk, i = 0;
 	    p < blk + blksize &&
 	    	 p + s->atr_ofs + s->atr_size <= blk + blksize &&
@@ -543,9 +554,101 @@ META_ORD meta_encode(META_SPC *s, char *blk, META_ORD blksize, META_AV *avlist,
 }
 
 
-/* Utility function for buildtree. Skips deleted items on the source list,
-   allocates a new item, makes it a copy of the current item on the source
-   list and advances the source list to the next item. The AV_FREE_P is
+#define BT_STKDEPTH	8
+
+/* Utility function for meta_buildtree.  This tries to find an encapsulation
+ * route for the space 'fromitem' resides in, to the space given in 'tospc'.
+ *
+ * It returns an array of encapsulating items, starting with the item that
+ * encapsulates 'fromitem's space and ending with an item that resides in the
+ * destination space. It returns the length of the route if a route was found, 
+ * and the route itself in the array given by encarr[encarrcnt].
+ *
+ * It finds routes by doing an iterative depth first search of items that can
+ * encapsulate a given space. It's not stricly depth first in that all siblings
+ * are visited before descending to children; but the search is not breadth
+ * first either, because grandchilden (and their respective offspring) are
+ * visited before nephews.
+ *
+ * On each level, encapsulating items that have the same vendor number as the
+ * 'fromitem' are tested first; however, this does not apply while descending.
+ *
+ * Searching the tree starting at the item to be encapsulated towards the
+ * ground space is more efficient than searching the tree from the ground
+ * space, simply because spaces generally have fewer items that can encapsulate
+ * them than spaces have items that can encapsulate subspaces. 
+ *
+ * In most cases this will just be a linear search, as most spaces have only
+ * one encapsulating item that can hold them. 
+ */
+
+static int findencroute(META_ITEM *fromitem, META_SPC *tospc, 
+		        META_ITEM **encarr, int encarrcnt)
+{
+	META_ITEM *ei, *i;
+	int sp;
+
+	ei = fromitem->spc->enc_items; 
+	sp = 0;
+	if (!ei) return 0;		    /* no encapsulation item for
+					       this space, then no route */
+	for(;;) {
+
+	    /* Search this level horizontally for a match, first for matching
+	     * space and vendor, then matching space only */
+
+	    for(i = ei; 
+		i && (i->spc != tospc || i->vnd != fromitem->vnd);
+		i = i->next);
+	    if (i) break;			    /* done */
+	    for(i = ei; 
+		i && (i->spc != tospc || i->vnd != META_VND_ANY);
+		i = i->next);
+	    if (i) break;			    /* done */
+
+	    /* The current space's list of encapsulating items does not contain
+	     * one that resides in the desired space. Push the first one at 
+	     * this level.
+	     */
+
+retry:
+	    if (ei->spc->enc_items) {
+
+		/* This sibling has children - push and descend */
+
+		if (sp >= encarrcnt - 1) return 0;	/* route too long */
+		encarr[sp++] = ei;
+		ei = ei->spc->enc_items;
+		continue;		/* ei is first one at new level, so
+					   do whole loop */
+	    }
+
+	    /* This sibling has no children - try next one */
+
+	    ei = ei->next;
+	    if (ei) goto retry;
+
+	    /* We've tested all our siblings and their children; pop parent
+	     * and continue there; if none, there's no route */
+
+	    if (!sp) return 0;				/* no more parents */
+	    ei = encarr[--sp];
+	    ei = ei->next;
+
+	    if (ei) goto retry;				
+	    return 0;					/* no more uncles */
+	}
+
+	/* We only get here after a route was found and the last item was i */
+
+	encarr[sp++] = i;
+	return sp;
+}
+
+
+/* Another utility function for buildtree. Skips deleted items on the source
+   list, allocates a new item, makes it a copy of the current item on the
+   source list and advances the source list to the next item. The AV_FREE_P is
    cleared on the copy, so that we don't double free things. */ 
 
 static META_AV *getcopy(META_AV **src)
@@ -657,7 +760,7 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 	META_SPC *dstspc;			/* current destination space */
 	META_AV *dsthead;			/* current destination head */
 	META_AV *newav;				/* encaps. pair to be added */
-	META_ITEM *encitem, *i;			/* current encapsulating item */
+	META_ITEM *encitem;			/* current encapsulating item */
 
 	/* Check parameters, init destination list and stack, get first item */
 
@@ -695,58 +798,33 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 			continue;
 		}
 
-		/* No, so try to reach the current space from this item's space
-		   by backtracking encapsulating items */
+		/* It does not, so try to find a route from the current item
+		 * to the current destination space, returning an array of
+		 * items starting with an item that encapsulates this item's
+		 * space and ending with an item that resides in the
+		 * destination space. */
 
-		for(encitem = srcav->i; encitem && encitem->spc != dstspc; ) {
+		itmsp = findencroute(srcav->i, dstspc, itmstk, BT_STKDEPTH);
+		if (!itmsp) {
 
-		    	/* Not there yet. First try to find an encapsulating 
-			   item for the current item's space that matches the 
-			   current item's vendor nr. */
-
-			for(i = encitem->spc->enc_items;
-			    i && i->vnd != srcav->i->vnd;
-			    i = i->next);
-
-			/* Couldn't match vendor, find enc. item allowing any */
-
-			if (!i) {
-				for(i = encitem->spc->enc_items;
-				    i && i->vnd != META_VND_ANY;
-				    i = i->next);
-			}
-
-			/* Add the current item's space's encapsulating item
-			 * that matches the current vendor as best as we can, 
-			 * to the encapsulating item stack */
-
-			encitem = i;
-			itmstk[itmsp++] = encitem;
-		}
-
-		/* If we didn't find an encapsulation route to the current
-		   space, go back one or more spaces on the destination stack
-		   and retry. */
-
-		if (!encitem) {
-
-			/* Empty stack of useless obtained items */
-
-			itmsp = 0;
-
-			/* If we're not at the bottom of the space stack yet, 
-			   pop destination list and space. */
+			/* If we didn not find an encapsulation route to the
+			 * current space, go back one or more spaces on the
+			 * destination stack and retry. */
 
 			if (dstsp) {
 
-				/* Go back one space and keep soing so if the
-				   space we get is a 'single attribute' space,
-				   because we know that there is already an
-				   attribute present from the fact that we're
-				   here -- a place which could only be reached
-				   using an encapsulating item. So all spaces
-				   below us already have their single allowed
-				   attribute present. */
+				/* We're not at the bottom of the space
+				 * stack yet, so pop destination list and
+				 * space. 
+				 * 
+				 * We keep doing so as long as the space we get
+				 * is a 'single attribute' space, because we
+				 * know that there is already an attribute
+				 * present from the fact that we're here -- a
+				 * place which could only be reached using an
+				 * encapsulating item. So all spaces below us
+				 * already have their single allowed attribute
+				 * present. */
 
 				do {
 					dstspc = dstspcstk[--dstsp];
@@ -756,8 +834,8 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 				while(dstspc->single && dstsp);
 
 				/* Try again to add the item, directly or via
-				   an encapsulation route, now we're at a lower
-				   space */
+				 * an encapsulation route, now we're at a lower
+				 * space */
 
 				continue;
 			}
@@ -780,9 +858,12 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 			continue;
 		}
 
-		/* We did succeed. Now follow the enc. route that we found,
-		   adding the encapsulating items to the destination list,
-		   while saving the destination list and space on its stack. */
+		/* We did succeed in finding an encapsulating route to the
+		 * current space. Now follow that route, starting at the end
+		 * (the encapsulating item in the current space), adding
+		 * encapsulating items until we reach the first one (which
+		 * encapsulates the the current item's space), saving each
+		 * each item's space on the destination stack. */
 
 		while(itmsp) {
 
@@ -792,11 +873,11 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 
 			encitem = itmstk[--itmsp];
 
-			/* If we have a destination head and we're in a fixed
-			 * field space, search the current destination list to
-			 * see if we already have the required encaps. item;
-			 * if so, add to its tail. This is all expensive, but
-			 * does not happen often. */
+			/* If we have a destination list head and we're in a
+			 * fixed field space, search the current destination
+			 * list to see if we already have the required encaps.
+			 * item; if so, add to its tail. This is all expensive,
+			 * but does not happen often. */
 
 			if (dsthead && dstspc->atr_size == 0) {
 				for(newav = dsthead;
