@@ -371,7 +371,7 @@ int vm_run(META *m, VM *vm)
 
 			/* Show source list */
 			if (msg_thresh[F_SEND] >= L_DEBUG) {
-				msg(F_SEND, L_DEBUG, "OP_DICTENC: encoding %s list:\n", R.ref.list == AVR_LIST_REP ? "REP" : "REQ");
+				msg(F_SEND, L_DEBUG, "OP_DICTENC: encoding %s list in space %s:\n", R.ref.list == AVR_LIST_REP ? "REP" : "REQ", R.i->spc->name);
 				meta_printavlist(m, vm->head[R.ref.list], 0);
 			}
 
