@@ -57,7 +57,7 @@ char eaptls_id[] = "EAPTLS - Copyright (C) 2005 Emile van Bergen.";
 #define MAX_CONVERSATIONS	128
 
 #define TLS_BUFFER_SIZE		32768	    /* Two TLS max. size records */
-#define TLS_KEYMAT_SIZE		64
+#define TLS_KEYMAT_SIZE		256
 
 #define RET_ERROR	    	0x40
 
@@ -74,8 +74,9 @@ char eaptls_id[] = "EAPTLS - Copyright (C) 2005 Emile van Bergen.";
 #define C_DI_TLS_ACTION		51	/* In C_DS_INTERNAL */
 #define C_DI_TLS_PAYLOAD	52	
 #define C_DI_TLS_KEYMAT		53
-#define C_DI_TLS_CLIENT_STATUS	54
-#define C_DI_TLS_CLIENT_SUBJECT	55
+#define C_DI_TLS_IVMAT		54
+#define C_DI_TLS_CLIENT_STATUS	55
+#define C_DI_TLS_CLIENT_SUBJECT	56
 
 #define C_DV_TLS_FLAGS_MORE	0x40
 #define C_DV_TLS_FLAGS_LENGTH	0x80
@@ -116,6 +117,9 @@ typedef struct conv {
     char keymatbuf[TLS_KEYMAT_SIZE];/* obtained after handshake */
     STR_T keymat;		    /* l initialized to desired size; p set
 				       when keying material is obtained */
+    char ivmatbuf[TLS_KEYMAT_SIZE]; /* obtained after handshake */
+    STR_T ivmat;		    /* l initialized to desired size; p set
+				       when keying material is obtained */
 } CONV_T;
 
 
@@ -132,11 +136,13 @@ typedef enum { RETURN, APP_GETKEY, APP_RECV, IA_RECV,
 
 /* command line arguments */
 
-META_ORD rx_space, tx_space, decpay_space;
-enum { DECPAY_RAW = 1, DECPAY_PEAPv2 = 2, DECPAY_DIAMETER = 4 } decpay_type;
-int wantclientcert;
+META_ORD rx_space = 0, tx_space = 0, decpay_space = 0;
+enum { DECPAY_RAW = 1, DECPAY_PEAPv2 = 2, DECPAY_DIAMETER = 4 } decpay_type = 0;
+int wantclientcert = 0; 
+int keymatlen = 64; 
+int ivmatlen = 0;
 STR_T keymat_label = cstr("client EAP encryption");
-int debug;
+int debug = 0;
 
 
 /* scratch buffer */
@@ -145,7 +151,7 @@ char buf[TLS_BUFFER_SIZE];
 
 
 /*
- * CALLBACKS (BRR)
+ * GnuTLS callbacks
  */
 
 
@@ -207,7 +213,11 @@ void tlslog(int level, const char *s)
 void usage()
 {
     fprintf(stderr, 
-"Usage: eaptls [-d] [-s attribute space base number]\n"
+"Usage: eaptls [-d] [-c] [-s transport attribute space base number]\n"
+"              [-tr] [-td] [-p ttls payload attribute space number]\n"
+"              [-l PRF label for deriving keying and IV material]\n"
+"              [-k length of keying material derived from MSK (64)]\n"
+"              [-i length of IV material (0)]\n"
 "       eaptls -v\n"
 "       eaptls -h\n");
     _exit(1);
@@ -220,7 +230,7 @@ void parseoptions(int argc, char **argv)
 
     /* Handle options */
 
-    while((c = getopt(argc, argv, "dhvs:p:t:l:c")) != -1) {
+    while((c = getopt(argc, argv, "dhvcs:p:t:l:k:i:")) != -1) {
         switch(c) {
 	  case 't': switch(*optarg) {
 			case 'r': decpay_type |= DECPAY_RAW; break;
@@ -249,6 +259,8 @@ void parseoptions(int argc, char **argv)
 	  case 'l': keymat_label.p = optarg;
 		    keymat_label.l = strlen(optarg);
 		    break;
+	  case 'k': keymatlen = atoi(optarg); break;
+	  case 'i': ivmatlen = atoi(optarg); break;
 	  case 'c': wantclientcert++; break;
           case 'd': debug++; break;
           case 'v': fprintf(stderr, "\nEAPTLS module " MODULE_VERSION ". "
@@ -513,17 +525,35 @@ int tls_state_machine(BUF_T *repbuf, CONV_T *conv, int app_action)
       /* Get keying material from handshake (and proceed with APP_RECV) */
 
       case APP_GETKEY:
-	conv->keymat.p = conv->keymatbuf;
-	gr = gnutls_prf(conv->session, keymat_label.l, keymat_label.p, 
-			0, 0, 0, conv->keymat.l, conv->keymat.p);
-	if (gr) {
-	    fprintf(stderr, "eaptls: gnutls_prf failed: "); gnutls_perror(gr);
-	    action = ERROR;
-	    break;
+	if (conv->keymat.l) {
+	    conv->keymat.p = conv->keymatbuf;
+	    gr = gnutls_prf(conv->session, keymat_label.l, keymat_label.p, 
+			    0, 0, 0, conv->keymat.l, conv->keymat.p);
+	    if (gr) {
+		fprintf(stderr, "eaptls: gnutls_prf failed: "); 
+		gnutls_perror(gr);
+		action = ERROR;
+		break;
+	    }
+	    if (debug) {
+		fprintf(stderr, "eaptls: Derived %d bytes of keying material using label [%.*s]:\n", conv->keymat.l, keymat_label.l, keymat_label.p);
+		hexdumpfd(2, conv->keymat.p, conv->keymat.l, 0);
+	    }
 	}
-	if (debug) {
-	    fprintf(stderr, "eaptls: generated %d bytes of keying material using label [%.*s]:\n", conv->keymat.l, keymat_label.l, keymat_label.p);
-	    hexdumpfd(2, conv->keymat.p, conv->keymat.l, 0);
+	if (conv->ivmat.l) {
+	    conv->ivmat.p = conv->ivmatbuf;
+	    gr = gnutls_prf(conv->session, keymat_label.l, keymat_label.p, 
+			    3, 0, 0, conv->ivmat.l, conv->ivmat.p);
+	    if (gr) {
+		fprintf(stderr, "eaptls: gnutls_prf failed: "); 
+		gnutls_perror(gr);
+		action = ERROR;
+		break;
+	    }
+	    if (debug) {
+		fprintf(stderr, "eaptls: Derived %d bytes of IV material using label [%.*s]:\n", conv->ivmat.l, keymat_label.l, keymat_label.p);
+		hexdumpfd(2, conv->ivmat.p, conv->ivmat.l, 0);
+	    }
 	}
 
 	/* fall through to APP_RECV */
@@ -795,7 +825,9 @@ int main(int argc, char **argv)
 	    conv->state = TLS_HANDSHAKE;
 	    conv->expiry = t + 60;
 	    conv->keymat.p = 0;
-	    conv->keymat.l = sizeof(conv->keymatbuf);
+	    conv->keymat.l = MIN(keymatlen, sizeof(conv->keymatbuf));
+	    conv->ivmat.p = 0;
+	    conv->ivmat.l = MIN(ivmatlen, sizeof(conv->ivmatbuf));
 	    conv->clientstatus = ~0;
 
 	    /* Create Gnutls session - may be resumed from existing one during
@@ -902,6 +934,17 @@ reply:
 		}
 	    }
 
+	    /* Add the IV material, if we have any */
+
+	    if (conv->ivmat.p) {
+		s = reply_startattr(&txbuf);
+		if (s.l >= conv->ivmat.l) {
+		    memcpy(s.p, conv->ivmat.p, conv->ivmat.l);
+		    reply_endattr(&txbuf, C_DS_INTERNAL, META_VND_ANY, 
+				  C_DI_TLS_IVMAT, conv->ivmat.l);
+		}
+	    }
+
 	    /* Add the client status, if we have any */
 
 	    if (conv->clientstatus != ~0)
